package ipv4

import (
	"fmt"
	"net"
	"time"

	"git.loafle.net/commons_go/logging"
	"git.loafle.net/commons_go/util/net/cidr"
	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"

	discoveryM "git.loafle.net/overflow/overflow_commons_go/modules/discovery/model"
	"git.loafle.net/overflow/overflow_discovery/commons/pcap"
)

func ScanHost(zone *discoveryM.Zone, dh *discoveryM.DiscoveryHost, resultChan chan interface{}, errChan chan error, stopChan chan struct{}) {
	ps, err := pcap.RetainScanner(zone)
	if nil != err {
		errChan <- fmt.Errorf("Discovery: Cannot retain pcap instance %v", err)
		return
	}

	defer func() {
		pcap.ReleaseScanner(zone)
	}()

	cr, err := cidr.NewCIDRRanger(zone.Network)
	if nil != err {
		errChan <- err
		return
	}

	hostRanges, err := getTargetHostRange(dh, cr)
	if nil != err {
		errChan <- err
		return
	}

	arpChan := ps.OpenARP()
	defer func() {
		ps.CloseARP(arpChan)
	}()

	go func() {
		hosts := make(map[string]*discoveryM.Host)
		for {
			select {
			case packet, ok := <-arpChan:
				if !ok {
					logging.Logger().Debugf("Discovery: arp channel is closed")
					return
				}
				if h := handlePacketARP(zone, hostRanges, hosts, packet); nil != h {
					resultChan <- h
				}
			case <-stopChan:
				return
			}
		}
	}()

	if err := sendARP(ps, zone, hostRanges, stopChan); nil != err {
		errChan <- err
		return
	}

	timer := time.NewTimer(10 * time.Second)

	select {
	case <-stopChan:
		return
	case <-timer.C:
		return
	}

}

func sendARP(ps pcap.PCapScanner, zone *discoveryM.Zone, hostRanges []net.IP, stopChan chan struct{}) error {
	hwAddr, err := net.ParseMAC(zone.Mac)
	if nil != err {
		return err
	}
	ip := net.ParseIP(zone.IP)
	if nil == ip {
		return fmt.Errorf("Discovery: IP(%s) of zone is not valid", zone.IP)
	}

	ethPacket := makePacketEthernet(hwAddr)
	arpPacket := makePacketARP(hwAddr, ip.To4())
	opts := gopacket.SerializeOptions{FixLengths: true, ComputeChecksums: true}
	buf := gopacket.NewSerializeBuffer()

	for _, targetHost := range hostRanges {
		arpPacket.DstProtAddress = []byte(targetHost)
		// log.Printf("ARP:%v", arpPacket)
		gopacket.SerializeLayers(buf, opts, &ethPacket, &arpPacket)
		if err := ps.WritePacketData(buf.Bytes()); err != nil {
			return err
		}

		timer := time.NewTimer(time.Microsecond * 100)

		select {
		case <-stopChan:
			return nil
		case <-timer.C:
		}

	}
	return nil
}

func handlePacketARP(zone *discoveryM.Zone, hostRanges []net.IP, hosts map[string]*discoveryM.Host, packet *layers.ARP) *discoveryM.Host {
	if packet.Operation != layers.ARPReply {
		return nil
	}

	ip := net.IP(packet.SourceProtAddress)
	if _, ok := hosts[ip.String()]; ok {
		return nil
	}

	inRange := false
	for _, h := range hostRanges {
		if h.Equal(ip) {
			inRange = true
			break
		}
	}
	if !inRange {
		return nil
	}

	h := &discoveryM.Host{}
	h.IP = ip.String()
	h.Mac = net.HardwareAddr(packet.SourceHwAddress).String()
	h.Zone = zone

	hosts[ip.String()] = h

	return h
}

func getTargetHostRange(dh *discoveryM.DiscoveryHost, cr cidr.CIDRRanger) ([]net.IP, error) {
	var firstIP net.IP
	if "" != dh.FirstScanRange {
		firstIP = net.ParseIP(dh.FirstScanRange)
		if nil == firstIP {
			return nil, fmt.Errorf("Discovery: IP(%v) of FirstScanRange host is not valid", firstIP)
		}
	}
	var lastIP net.IP
	if "" != dh.LastScanRange {
		lastIP = net.ParseIP(dh.LastScanRange)
		if nil == lastIP {
			return nil, fmt.Errorf("Discovery: IP(%v) of LastScanRange host is not valid", lastIP)
		}
	}

	includeIPs := make([]net.IP, 0)
	for _, iHost := range dh.IncludeHosts {
		iIP := net.ParseIP(iHost)
		if nil == iIP {
			return nil, fmt.Errorf("Discovery: IP(%v) of include host is not valid", iHost)
		}
		includeIPs = append(includeIPs, iIP)
	}

	excludeIPs := make([]net.IP, 0)
	for _, eHost := range dh.ExcludeHosts {
		eIP := net.ParseIP(eHost)
		if nil == eIP {
			return nil, fmt.Errorf("Discovery: IP(%v) of exclude host is not valid", eHost)
		}
		excludeIPs = append(excludeIPs, eIP)
	}

	ranges, err := cr.Ranges(firstIP, lastIP, includeIPs, excludeIPs)
	if nil != err {
		return nil, err
	}

	return ranges, nil
}

func makePacketEthernet(hw net.HardwareAddr) layers.Ethernet {
	return layers.Ethernet{
		SrcMAC:       hw,
		DstMAC:       net.HardwareAddr{0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
		EthernetType: layers.EthernetTypeARP,
	}
}

func makePacketARP(hw net.HardwareAddr, ip net.IP) layers.ARP {
	return layers.ARP{
		AddrType:          layers.LinkTypeEthernet,
		Protocol:          layers.EthernetTypeIPv4,
		HwAddressSize:     6,
		ProtAddressSize:   4,
		Operation:         layers.ARPRequest,
		SourceHwAddress:   []byte(hw),
		SourceProtAddress: []byte(ip),
		DstHwAddress:      []byte{0, 0, 0, 0, 0, 0},
	}
}
